import itertools
import logging
from time import perf_counter

import h5py
import numpy as np
import pandas as pd
from collections.abc import Iterable
from typing import Dict

from utils_find_1st import find_1st, cmp_equal

from aliby.io.base import BridgeH5
from aliby.utils import timed


#################### Dynamic version ##################################


def load_attributes(file: str, group="/"):
    with h5py.File(file, "r") as f:
        meta = dict(f[group].attrs.items())
    return meta


class DynamicWriter:
    data_types = {}
    group = ""
    compression = None

    def __init__(self, file: str):
        self.file = file
        self.metadata = load_attributes(file)

    def _append(self, data, key, hgroup):
        """Append data to existing dataset."""
        try:
            n = len(data)
        except:
            # Attributes have no length
            n = 1
        if key not in hgroup:
            # TODO Include sparsity check
            max_shape, dtype = self.datatypes[key]
            shape = (n,) + max_shape[1:]
            hgroup.create_dataset(
                key,
                shape=shape,
                maxshape=max_shape,
                dtype=dtype,
                compression=self.compression,
            )
            hgroup[key][()] = data
        else:
            # The dataset already exists, expand it

            try:  # FIXME This is broken by bugged mother-bud assignment
                dset = hgroup[key]
                dset.resize(dset.shape[0] + n, axis=0)
                dset[-n:] = data
            except:
                logging.debug(
                    "DynamicWriter:Inconsistency between dataset shape and new empty data"
                )
        return

    def _overwrite(self, data, key, hgroup):
        """Overwrite existing dataset with new data"""
        # We do not append to mother_assign; raise error if already saved
        n = len(data)
        max_shape, dtype = self.datatypes[key]
        if key in hgroup:
            del hgroup[key]
        hgroup.require_dataset(
            key, shape=(n,), dtype=dtype, compression=self.compression
        )
        hgroup[key][()] = data

    def _check_key(self, key):
        if key not in self.datatypes:
            raise KeyError(f"No defined data type for key {key}")

    def write(self, data, overwrite: list):
        # Data is a dictionary, if not, make it one
        # Overwrite data is a dictionary
        with h5py.File(self.file, "a") as store:
            hgroup = store.require_group(self.group)

            for key, value in data.items():
                # We're only saving data that has a pre-defined data-type
                self._check_key(key)
                try:
                    if key.startswith("attrs/"):  # metadata
                        key = key.split("/")[1]  # First thing after attrs
                        hgroup.attrs[key] = value
                    elif key in overwrite:
                        self._overwrite(value, key, hgroup)
                    else:
                        self._append(value, key, hgroup)
                except Exception as e:
                    print(key, value)
                    raise (e)
        return


##################### Special instances #####################
class TilerWriter(DynamicWriter):
    datatypes = {
        "trap_locations": ((None, 2), np.uint16),
        "drifts": ((None, 2), np.float32),
        "attrs/tile_size": ((1,), np.uint16),
        "attrs/max_size": ((1,), np.uint16),
    }
    group = "trap_info"


tile_size = 117


@timed()
def save_complex(array, dataset):
    # Dataset needs to be 2D
    n = len(array)
    if n > 0:
        dataset.resize(dataset.shape[0] + n, axis=0)
        dataset[-n:, 0] = array.real
        dataset[-n:, 1] = array.imag


@timed()
def load_complex(dataset):
    array = dataset[:, 0] + 1j * dataset[:, 1]
    return array


class BabyWriter(DynamicWriter):
    compression = "gzip"
    max_ncells = 2e5  # Could just make this None
    max_tps = 1e3  # Could just make this None
    chunk_cells = 25  # The number of cells in a chunk for edge masks
    default_tile_size = 117
    datatypes = {
        "centres": ((None, 2), np.uint16),
        "position": ((None,), np.uint16),
        "angles": ((None,), h5py.vlen_dtype(np.float32)),
        "radii": ((None,), h5py.vlen_dtype(np.float32)),
        "edgemasks": ((max_ncells, max_tps, tile_size, tile_size), np.bool),
        "ellipse_dims": ((None, 2), np.float32),
        "cell_label": ((None,), np.uint16),
        "trap": ((None,), np.uint16),
        "timepoint": ((None,), np.uint16),
        "mother_assign": ((None,), h5py.vlen_dtype(np.uint16)),
        "mother_assign_dynamic": ((None,), np.uint16),
        "volumes": ((None,), np.float32),
    }
    group = "cell_info"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Get max_tps and trap info
        self._traps_initialised = False

    def __init_trap_info(self):
        # Should only be run after the traps have been initialised
        trap_metadata = load_attributes(self.file, "trap_info")
        tile_size = trap_metadata.get("tile_size", self.default_tile_size)
        max_tps = self.metadata["time_settings/ntimepoints"][0]
        self.datatypes["edgemasks"] = (
            (self.max_ncells, max_tps, tile_size, tile_size),
            np.bool,
        )
        self._traps_initialised = True

    def __init_edgemasks(self, hgroup, edgemasks, current_indices, n_cells):
        # Create values dataset
        # This holds the edge masks directly and
        # Is of shape (n_tps, n_cells, tile_size, tile_size)
        key = "edgemasks"
        max_shape, dtype = self.datatypes[key]
        shape = (n_cells, 1) + max_shape[2:]
        chunks = (self.chunk_cells, 1) + max_shape[2:]
        val_dset = hgroup.create_dataset(
            "values",
            shape=shape,
            maxshape=max_shape,
            dtype=dtype,
            chunks=chunks,
            compression=self.compression,
        )
        val_dset[:, 0] = edgemasks
        # Create index dataset
        # Holds the (trap, cell_id) description used to index into the
        # values and is of shape (n_cells, 2)
        ix_max_shape = (max_shape[0], 2)
        ix_shape = (0, 2)
        ix_dtype = np.uint16
        ix_dset = hgroup.create_dataset(
            "indices",
            shape=ix_shape,
            maxshape=ix_max_shape,
            dtype=ix_dtype,
            compression=self.compression,
        )
        save_complex(current_indices, ix_dset)

    def __append_edgemasks(self, hgroup, edgemasks, current_indices):
        key = "edgemasks"
        val_dset = hgroup["values"]
        ix_dset = hgroup["indices"]
        existing_indices = load_complex(ix_dset)
        # Check if there are any new labels
        available = np.in1d(current_indices, existing_indices)
        missing = current_indices[~available]
        all_indices = np.concatenate([existing_indices, missing])
        # Resizing
        t = perf_counter()
        n_tps = val_dset.shape[1] + 1
        n_add_cells = len(missing)
        # RESIZE DATASET FOR TIME and Cells
        new_shape = (val_dset.shape[0] + n_add_cells, n_tps) + val_dset.shape[2:]
        val_dset.resize(new_shape)
        logging.debug(f"Timing:resizing:{perf_counter() - t}")
        # Writing data
        cell_indices = np.where(np.in1d(all_indices, current_indices))[0]
        for ix, mask in zip(cell_indices, edgemasks):
            try:
                val_dset[ix, n_tps - 1] = mask
            except Exception as e:
                logging.debug(f"{ix}, {n_tps}, {val_dset.shape}")
        # Save the index values
        save_complex(missing, ix_dset)

    def write_edgemasks(self, data, keys, hgroup):
        if not self._traps_initialised:
            self.__init_trap_info()
        # DATA is TRAP_IDS, CELL_LABELS, EDGEMASKS in a structured array
        key = "edgemasks"
        val_key = "values"
        idx_key = "indices"
        # Length of edgemasks
        traps, cell_labels, edgemasks = data
        n_cells = len(cell_labels)
        hgroup = hgroup.require_group(key)
        current_indices = np.array(traps) + 1j * np.array(cell_labels)
        if val_key not in hgroup:
            self.__init_edgemasks(hgroup, edgemasks, current_indices, n_cells)
        else:
            self.__append_edgemasks(hgroup, edgemasks, current_indices)

    def write(self, data, overwrite: list):
        with h5py.File(self.file, "a") as store:
            hgroup = store.require_group(self.group)

            for key, value in data.items():
                # We're only saving data that has a pre-defined data-type
                self._check_key(key)
                try:
                    if key.startswith("attrs/"):  # metadata
                        key = key.split("/")[1]  # First thing after attrs
                        hgroup.attrs[key] = value
                    elif key in overwrite:
                        self._overwrite(value, key, hgroup)
                    elif key == "edgemasks":
                        keys = ["trap", "cell_label", "edgemasks"]
                        value = [data[x] for x in keys]
                        self.write_edgemasks(value, keys, hgroup)
                    else:
                        self._append(value, key, hgroup)
                except Exception as e:
                    print(key, value)
                    raise (e)
        return


#################### Extraction version ###############################
class Writer(BridgeH5):
    """
    Class in charge of transforming data into compatible formats

    Decoupling interface from implementation!

    Parameters
    ----------
        filename: str Name of file to write into
        flag: str, default=None
            Flag to pass to the default file reader. If None the file remains closed.
        compression: str, default=None
            Compression method passed on to h5py writing functions (only used for
        dataframes and other array-like data.)
    """

    def __init__(self, filename, compression=None):
        super().__init__(filename, flag=None)

        if compression is None:
            self.compression = "gzip"

    def write(
        self,
        path: str,
        data: Iterable = None,
        meta: Dict = {},
        overwrite: str = None,
    ):
        """
        Parameters
        ----------
        path : str
            Path inside h5 file to write into.
        data : Iterable, default = None
        meta : Dict, default = {}


        """
        self.id_cache = {}
        with h5py.File(self.filename, "a") as f:
            if overwrite == "overwrite":  # TODO refactor overwriting
                if path in f:
                    del f[path]
            elif overwrite == "accumulate":  # Add a number if needed
                if path in f:
                    parent, name = path.rsplit("/", maxsplit=1)
                    n = sum([x.startswith(name) for x in f[path]])
                    path = path + str(n).zfill(3)
            elif overwrite == "skip":
                if path in f:
                    logging.debug("Skipping dataset {}".format(path))

            logging.debug(
                "{} {} to {} and {} metadata fields".format(
                    overwrite, type(data), path, len(meta)
                )
            )
            if data is not None:
                self.write_dset(f, path, data)
            if meta:
                for attr, metadata in meta.items():
                    self.write_meta(f, path, attr, data=metadata)

    def write_dset(self, f: h5py.File, path: str, data: Iterable):
        if isinstance(data, pd.DataFrame):
            self.write_pd(f, path, data, compression=self.compression)
        elif isinstance(data, pd.MultiIndex):
            self.write_index(f, path, data)  # , compression=self.compression)
        elif isinstance(data, Iterable):
            self.write_arraylike(f, path, data)
        else:
            self.write_atomic(data, f, path)

    def write_meta(self, f: h5py.File, path: str, attr: str, data: Iterable):
        obj = f.require_group(path)

        obj.attrs[attr] = data

    @staticmethod
    def write_arraylike(f: h5py.File, path: str, data: Iterable, **kwargs):
        if path in f:
            del f[path]

        narray = np.array(data)

        chunks = None
        if narray.any():
            chunks = (1, *narray.shape[1:])

        dset = f.create_dataset(
            path,
            shape=narray.shape,
            chunks=chunks,
            dtype="int",
            compression=kwargs.get("compression", None),
        )
        dset[()] = narray

    @staticmethod  # TODO Use this function to implement Diane's dynamic writer
    def write_dynamic(f: h5py.File, path: str, data: Iterable):
        pass

    @staticmethod
    def write_index(f, path, pd_index, **kwargs):
        f.require_group(path)  # TODO check if we can remove this
        for i, name in enumerate(pd_index.names):
            ids = pd_index.get_level_values(i)
            id_path = path + "/" + name
            f.create_dataset(
                name=id_path,
                shape=(len(ids),),
                dtype="uint16",
                compression=kwargs.get("compression", None),
            )
            indices = f[id_path]
            indices[()] = ids

    def write_pd(self, f, path, df, **kwargs):
        values_path = path + "values" if path.endswith("/") else path + "/values"
        if path not in f:
            max_ncells = 2e5

            max_tps = 1e3
            f.create_dataset(
                name=values_path,
                shape=df.shape,
                # chunks=(min(df.shape[0], 1), df.shape[1]),
                # dtype=df.dtypes.iloc[0], This is making NaN in ints into negative vals
                dtype="float",
                maxshape=(max_ncells, max_tps),
                compression=kwargs.get("compression", None),
            )
            dset = f[values_path]
            dset[()] = df.values

            for name in df.index.names:
                indices_path = "/".join((path, name))
                f.create_dataset(
                    name=indices_path,
                    shape=(len(df),),
                    dtype="uint16",  # Assuming we'll always use int indices
                    chunks=True,
                    maxshape=(max_ncells,),
                )
                dset = f[indices_path]
                dset[()] = df.index.get_level_values(level=name).tolist()

            if df.columns.dtype == np.int or df.columns.dtype == np.dtype("uint"):
                tp_path = path + "/timepoint"
                f.create_dataset(
                    name=tp_path,
                    shape=(df.shape[1],),
                    maxshape=(max_tps,),
                    dtype="uint16",
                )
                tps = df.columns.tolist()
                f[tp_path][tps] = tps
            else:
                f[path].attrs["columns"] = df.columns.tolist()
        else:
            dset = f[values_path]

            # Filter out repeated timepoints
            new_tps = set(df.columns)
            if path + "/timepoint" in f:
                new_tps = new_tps.difference(f[path + "/timepoint"][()])
            df = df[new_tps]

            if (
                not hasattr(self, "id_cache") or not df.index.nlevels in self.id_cache
            ):  # Use cache dict to store previously-obtained indices
                self.id_cache[df.index.nlevels] = {}
                existing_ids = self.get_existing_ids(
                    f, [path + "/" + x for x in df.index.names]
                )
                # Split indices in existing and additional
                new = df.index.tolist()
                if df.index.nlevels == 1:  # Cover for cases with a single index
                    new = [(x,) for x in df.index.tolist()]
                (
                    found_multis,
                    self.id_cache[df.index.nlevels]["additional_multis"],
                ) = self.find_ids(
                    existing=existing_ids,
                    new=new,
                )
                found_indices = np.array(locate_indices(existing_ids, found_multis))

                # We must sort our indices for h5py indexing
                incremental_existing = np.argsort(found_indices)
                self.id_cache[df.index.nlevels]["found_indices"] = found_indices[
                    incremental_existing
                ]
                self.id_cache[df.index.nlevels]["found_multi"] = found_multis[
                    incremental_existing
                ]

            existing_values = df.loc[
                [
                    _tuple_or_int(x)
                    for x in self.id_cache[df.index.nlevels]["found_multi"]
                ]
            ].values
            new_values = df.loc[
                [
                    _tuple_or_int(x)
                    for x in self.id_cache[df.index.nlevels]["additional_multis"]
                ]
            ].values
            ncells, ntps = f[values_path].shape

            # Add found cells
            dset.resize(dset.shape[1] + df.shape[1], axis=1)
            dset[:, ntps:] = np.nan
            for i, tp in enumerate(df.columns):
                dset[
                    self.id_cache[df.index.nlevels]["found_indices"], tp
                ] = existing_values[:, i]
            # Add new cells
            n_newcells = len(self.id_cache[df.index.nlevels]["additional_multis"])
            dset.resize(dset.shape[0] + n_newcells, axis=0)
            dset[ncells:, :] = np.nan

            for i, tp in enumerate(df.columns):
                dset[ncells:, tp] = new_values[:, i]

            # save indices
            for i, name in enumerate(df.index.names):
                tmp = path + "/" + name
                dset = f[tmp]
                n = dset.shape[0]
                dset.resize(n + n_newcells, axis=0)
                dset[n:] = (
                    self.id_cache[df.index.nlevels]["additional_multis"][:, i]
                    if len(self.id_cache[df.index.nlevels]["additional_multis"].shape)
                    > 1
                    else self.id_cache[df.index.nlevels]["additional_multis"]
                )

            tmp = path + "/timepoint"
            dset = f[tmp]
            n = dset.shape[0]
            dset.resize(n + df.shape[1], axis=0)
            dset[n:] = df.columns.tolist()

    @staticmethod
    def get_existing_ids(f, paths):
        # Fetch indices and convert them to a (nentries, nlevels) ndarray
        return np.array([f[path][()] for path in paths]).T

    @staticmethod
    def find_ids(existing, new):
        # Compare two tuple sets and return the intersection and difference
        # (elements in the 'new' set not in 'existing')
        set_existing = set([tuple(*x) for x in zip(existing.tolist())])
        existing_cells = np.array(list(set_existing.intersection(new)))
        new_cells = np.array(list(set(new).difference(set_existing)))

        return (
            existing_cells,
            new_cells,
        )


# @staticmethod
def locate_indices(existing, new):
    if new.any():
        if new.shape[1] > 1:
            return [
                find_1st(
                    (existing[:, 0] == n[0]) & (existing[:, 1] == n[1]), True, cmp_equal
                )
                for n in new
            ]
        else:
            return [find_1st(existing[:, 0] == n, True, cmp_equal) for n in new]
    else:
        return []


# def tuple_or_int(x):
#     if isinstance(x, Iterable):
#         return tuple(x)
#     else:
#         return x
def _tuple_or_int(x):
    # Convert tuple to int if it only contains one value
    if len(x) == 1:
        return x[0]
    else:
        return x



"""
Testing the "run" functions in the pipeline elements.
"""
import pytest
pytest.mark.skip(reason='All tests still WIP')

# Todo: data needed: an experiment object
# Todo: data needed: an sqlite database
# Todo: data needed: a Shelf storage
class TestPipeline:
    def test_experiment(self):
        pass

    def test_omero_experiment(self):
        pass

    def test_tiler(self):
        pass

    def test_baby_client(self):
        pass

    def test_baby_runner(self):
        pass

    def test_pipeline(self):
        pass




import numpy as np
from time import perf_counter
from pathlib import Path

import tensorflow as tf

from aliby.io.writer import DynamicWriter


def initialise_tf(version):
    # Initialise tensorflow
    if version == 1:
        core_config = tf.ConfigProto()
        core_config.gpu_options.allow_growth = True
        session = tf.Session(config=core_config)
        return session
    # TODO this only works for TF2
    if version == 2:
        gpus = tf.config.experimental.list_physical_devices("GPU")
        if gpus:
            for gpu in gpus:
                tf.config.experimental.set_memory_growth(gpu, True)
            logical_gpus = tf.config.experimental.list_logical_devices("GPU")
            print(len(gpus), "Physical GPUs,", len(logical_gpus), "Logical GPUs")
        return None


def timer(func, *args, **kwargs):
    start = perf_counter()
    result = func(*args, **kwargs)
    print(f"Function {func.__name__}: {perf_counter() - start}s")
    return result


################## CUSTOM OBJECTS ##################################


class ModelPredictor:
    """Generic object that takes a NN and returns the prediction.

    Use for predicting fluorescence/other from bright field.
    This does not do instance segmentations of anything.
    """

    def __init__(self, tiler, model, name):
        self.tiler = tiler
        self.model = model
        self.name = name

    def get_data(self, tp):
        # Change axes to X,Y,Z rather than Z,Y,X
        return self.tiler.get_tp_data(tp, self.bf_channel).swapaxes(1, 3).swapaxes(1, 2)

    def format_result(self, result, tp):
        return {self.name: result, "timepoints": [tp] * len(result)}

    def run_tp(self, tp, **kwargs):
        """Simulating processing time with sleep"""
        # Access the image
        segmentation = self.model.predict(self.get_data(tp))
        return self._format_result(segmentation, tp)


class ModelPredictorWriter(DynamicWriter):
    def __init__(self, file, name, shape, dtype):
        super.__init__(file)
        self.datatypes = {name: (shape, dtype), "timepoint": ((None,), np.uint16)}
        self.group = f"{self.name}_info"


class Saver:
    channel_names = {0: "BrightField", 1: "GFP"}

    def __init__(self, tiler, save_directory, pos_name):
        """This class straight up saves the trap data for use with neural networks in the future."""
        self.tiler = tiler
        self.name = pos_name
        self.save_dir = Path(save_directory)

    def channel_dir(self, index):
        ch_dir = self.save_dir / self.channel_names[index]
        if not ch_dir.exists():
            ch_dir.mkdir()
        return ch_dir

    def get_data(self, tp, ch):
        return self.tiler.get_tp_data(tp, ch).swapaxes(1, 3).swapaxes(1, 2)

    def cache(self, tp):
        # Get a given time point
        # split into channels
        for ch in self.channel_names:
            ch_dir = self.channel_dir(ch)
            data = self.get_data(tp, ch)
            for tid, trap in enumerate(data):
                np.save(ch_dir / f"{self.name}_{tid}_{tp}.npy", trap)
        return



import logging
from pathlib import Path, PosixPath
from time import perf_counter
from typing import Union
from itertools import groupby
from collections.abc import Iterable

from utils_find_1st import find_1st, cmp_equal
import h5py
import numpy as np
from scipy import ndimage
from scipy.sparse.base import isdense

from aliby.io.matlab import matObject
from aliby.utils import timed
from aliby.io.writer import load_complex


def cell_factory(store, type="matlab"):
    if isinstance(store, matObject):
        return CellsMat(store)
    if type == "matlab":
        mat_object = matObject(store)
        return CellsMat(mat_object)
    elif type == "hdf5":
        return CellsHDF(store)
    else:
        raise TypeError(
            "Could not get cells for type {}:" "valid types are matlab and hdf5"
        )


class Cells:
    """An object that gathers information about all the cells in a given
    trap.
    This is the abstract object, used for type testing
    """

    def __init__(self):
        pass

    @staticmethod
    def from_source(source: Union[PosixPath, str], kind: str = None):
        if isinstance(source, str):
            source = Path(source)
        if kind is None:  # Infer kind from filename
            kind = "matlab" if source.suffix == ".mat" else "hdf5"
        return cell_factory(source, kind)

    @staticmethod
    def _asdense(array):
        if not isdense(array):
            array = array.todense()
        return array

    @staticmethod
    def _astype(array, kind):
        # Convert sparse arrays if needed and if kind is 'mask' it fills the outline
        array = Cells._asdense(array)
        if kind == "mask":
            array = ndimage.binary_fill_holes(array).astype(int)
        return array

    @classmethod
    def hdf(cls, fpath):
        return CellsHDF(fpath)

    @classmethod
    def mat(cls, path):
        return CellsMat(matObject(store))


class CellsHDF(Cells):
    def __init__(self, filename, path="cell_info"):
        self.filename = filename
        self.cinfo_path = path
        self._edgem_indices = None
        self._edgemasks = None
        self._tile_size = None

    def __getitem__(self, item):
        if item == "edgemasks":
            return self.edgemasks
        _item = "_" + item
        if not hasattr(self, _item):
            setattr(self, _item, self._fetch(item))
        return getattr(self, _item)

    def _get_idx(self, cell_id, trap_id):
        return (self["cell_label"] == cell_id) & (self["trap"] == trap_id)

    def _fetch(self, path):
        with h5py.File(self.filename, mode="r") as f:
            return f[self.cinfo_path][path][()]

    @property
    def ntraps(self):
        with h5py.File(self.filename, mode="r") as f:
            return len(f["/trap_info/trap_locations"][()])

    @property
    def traps(self):
        return list(set(self["trap"]))

    @property
    def tile_size(self):  # TODO read from metadata
        if self._tile_size is None:
            with h5py.File(self.filename, mode="r") as f:
                self._tile_size == f["trap_info/tile_size"][0]
        return self._tile_size

    @property
    def edgem_indices(self):
        if self._edgem_indices is None:
            edgem_path = "edgemasks/indices"
            self._edgem_indices = load_complex(self._fetch(edgem_path))
        return self._edgem_indices

    @property
    def edgemasks(self):
        if self._edgemasks is None:
            edgem_path = "edgemasks/values"
            self._edgemasks = self._fetch(edgem_path)

        return self._edgemasks

    def _edgem_where(self, cell_id, trap_id):
        ix = trap_id + 1j * cell_id
        return find_1st(self.edgem_indices == ix, True, cmp_equal)

    @property
    def labels(self):
        """
        Return all cell labels in object
        We use mother_assign to list traps because it is the only propriety that appears even
        when no cells are found"""
        return [self.labels_in_trap(trap) for trap in self.traps]

    def where(self, cell_id, trap_id):
        """
        Returns
        Parameters
        ----------
            cell_id: int
                Cell index
            trap_id: int
                Trap index

        Returns
        ----------
            indices int array
            boolean mask array
            edge_ix int array
        """
        indices = self._get_idx(cell_id, trap_id)
        edgem_ix = self._edgem_where(cell_id, trap_id)
        return (
            self["timepoint"][indices],
            indices,
            edgem_ix,
        )  # FIXME edgem_ix makes output different to matlab's Cell

    def outline(self, cell_id, trap_id):
        times, indices, cell_ix = self.where(cell_id, trap_id)
        return times, self["edgemasks"][cell_ix, times]

    def mask(self, cell_id, trap_id):
        times, outlines = self.outline(cell_id, trap_id)
        return times, np.array(
            [ndimage.morphology.binary_fill_holes(o) for o in outlines]
        )

    def at_time(self, timepoint, kind="mask"):
        ix = self["timepoint"] == timepoint
        cell_ix = self["cell_label"][ix]
        traps = self["trap"][ix]
        indices = traps + 1j * cell_ix
        choose = np.in1d(self.edgem_indices, indices)
        edgemasks = self["edgemasks"][choose, timepoint]
        masks = [
            self._astype(edgemask, kind) for edgemask in edgemasks if edgemask.any()
        ]
        return self.group_by_traps(traps, masks)

    def group_by_traps(self, traps, data):
        # returns a dict with traps as keys and labels as value
        iterator = groupby(zip(traps, data), lambda x: x[0])
        d = {key: [x[1] for x in group] for key, group in iterator}
        d = {i: d.get(i, []) for i in self.traps}
        return d

    def labels_in_trap(self, trap_id):
        # Return set of cell ids in a trap.
        return set((self["cell_label"][self["trap"] == trap_id]))

    def labels_at_time(self, timepoint):
        labels = self["cell_label"][self["timepoint"] == timepoint]
        traps = self["trap"][self["timepoint"] == timepoint]
        return self.group_by_traps(traps, labels)


class CellsMat(Cells):
    def __init__(self, mat_object):
        super(CellsMat, self).__init__()
        # TODO add __contains__ to the matObject
        timelapse_traps = mat_object.get(
            "timelapseTrapsOmero", mat_object.get("timelapseTraps", None)
        )
        if timelapse_traps is None:
            raise NotImplementedError(
                "Could not find a timelapseTraps or "
                "timelapseTrapsOmero object. Cells "
                "from cellResults not implemented"
            )
        else:
            self.trap_info = timelapse_traps["cTimepoint"]["trapInfo"]

            if isinstance(self.trap_info, list):
                self.trap_info = {
                    k: list([res.get(k, []) for res in self.trap_info])
                    for k in self.trap_info[0].keys()
                }

    def where(self, cell_id, trap_id):
        times, indices = zip(
            *[
                (tp, np.where(cell_id == x)[0][0])
                for tp, x in enumerate(self.trap_info["cellLabel"][:, trap_id].tolist())
                if np.any(cell_id == x)
            ]
        )
        return times, indices

    def outline(self, cell_id, trap_id):
        times, indices = self.where(cell_id, trap_id)
        info = self.trap_info["cell"][times, trap_id]

        def get_segmented(cell, index):
            if cell["segmented"].ndim == 0:
                return cell["segmented"][()].todense()
            else:
                return cell["segmented"][index].todense()

        segmentation_outline = [
            get_segmented(cell, idx) for idx, cell in zip(indices, info)
        ]
        return times, np.array(segmentation_outline)

    def mask(self, cell_id, trap_id):
        times, outlines = self.outline(cell_id, trap_id)
        return times, np.array(
            [ndimage.morphology.binary_fill_holes(o) for o in outlines]
        )

    def at_time(self, timepoint, kind="outline"):

        """Returns the segmentations for all the cells at a given timepoint.

        FIXME: this is extremely hacky and accounts for differently saved
            results in the matlab object. Deprecate ASAP.
        """
        # Case 1: only one cell per trap: trap_info['cell'][timepoint] is a
        # structured array
        if isinstance(self.trap_info["cell"][timepoint], dict):
            segmentations = [
                self._astype(x, "outline")
                for x in self.trap_info["cell"][timepoint]["segmented"]
            ]
        # Case 2: Multiple cells per trap: it becomes a list of arrays or
        # dictionaries,  one for each trap
        # Case 2.1 : it's a dictionary
        elif isinstance(self.trap_info["cell"][timepoint][0], dict):
            segmentations = []
            for x in self.trap_info["cell"][timepoint]:
                seg = x["segmented"]
                if not isinstance(seg, np.ndarray):
                    seg = [seg]
                segmentations.append([self._astype(y, "outline") for y in seg])
        # Case 2.2 : it's an array
        else:
            segmentations = [
                [self._astype(y, type) for y in x["segmented"]] if x.ndim != 0 else []
                for x in self.trap_info["cell"][timepoint]
            ]
            # Return dict for compatibility with hdf5 output
        return {i: v for i, v in enumerate(segmentations)}

    def labels_at_time(self, tp):
        labels = self.trap_info["cellLabel"]
        labels = [_aslist(x) for x in labels[tp]]
        labels = {i: [lbl for lbl in lblset] for i, lblset in enumerate(labels)}
        return labels

    @property
    def ntraps(self):
        return len(self.trap_info["cellLabel"][0])

    @property
    def tile_size(self):
        pass


class ExtractionRunner:
    """An object to run extraction of fluorescence, and general data out of
    segmented data.

    Configure with what extraction we want to run.
    Cell selection criteria.
    Filtering criteria.
    """

    def __init__(self, tiler, cells):
        pass

    def run(self, keys, store, **kwargs):
        pass


def _aslist(x):
    if isinstance(x, Iterable):
        if hasattr(x, "tolist"):
            x = x.tolist()
    else:
        x = [x]
    return x



import re
import struct


def clean_ascii(text):
    return re.sub(r'[^\x20-\x7F]', '.', text)


def xxd(x, start=0, stop=None):
    if stop is None:
        stop = len(x)
    for i in range(start, stop, 8):
        # Row number
        print("%04d" % i, end="   ")
        # Hexadecimal bytes
        for r in range(i, i + 8):
            print("%02x" % x[r], end="")
            if (r + 1) % 4 == 0:
                print("  ", end="")
        # ASCII
        print("   ", clean_ascii(x[i:i + 8].decode('utf-8', errors='ignore')),
              "   ", end="")
        # Int32
        print('{:>10} {:>10}'.format(*struct.unpack('II', x[i: i + 8])),
              end="   ")
        print("")  # Newline
    return


# Buffer reading functions
def read_int(buffer, n=1):
    res = struct.unpack('I' * n, buffer.read(4 * n))
    if n == 1:
        res = res[0]
    return res


def read_string(buffer):
    return ''.join([x.decode() for x in iter(lambda: buffer.read(1), b'\x00')])


def read_delim(buffer, n):
    delim = read_int(buffer, n)
    assert all([x == 0 for x in delim]), "Unknown nonzero value in delimiter"



"""
Parse microscopy log files according to specified JSON grammars.
Produces dictionary to include in HDF5
"""
import glob
import os
import numpy as np
import pandas as pd
from datetime import datetime, timezone
from pytz import timezone

from logfile_parser import Parser

# Paradigm: able to do something with all datatypes present in log files,
# then pare down on what specific information is really useful later.

# Needed because HDF5 attributes do not support dictionaries
def flatten_dict(nested_dict, separator="/"):
    """
    Flattens nested dictionary
    """
    df = pd.json_normalize(nested_dict, sep=separator)
    return df.to_dict(orient="records")[0]


# Needed because HDF5 attributes do not support datetime objects
# Takes care of time zones & daylight saving
def datetime_to_timestamp(time, locale="Europe/London"):
    """
    Convert datetime object to UNIX timestamp
    """
    return timezone(locale).localize(time).timestamp()


def find_file(root_dir, regex):
    file = glob.glob(os.path.join(str(root_dir), regex))
    if len(file) != 1:
        return None
    else:
        return file[0]


# TODO: re-write this as a class if appropriate
# WARNING: grammars depend on the directory structure of a locally installed
# logfile_parser repo
def parse_logfiles(
    root_dir,
    acq_grammar="multiDGUI_acq_format.json",
    log_grammar="multiDGUI_log_format.json",
):
    """
    Parse acq and log files depending on the grammar specified, then merge into
    single dict.
    """
    # Both acq and log files contain useful information.
    # ACQ_FILE = 'flavin_htb2_glucose_long_ramp_DelftAcq.txt'
    # LOG_FILE = 'flavin_htb2_glucose_long_ramp_Delftlog.txt'
    log_parser = Parser(log_grammar)
    try:
        log_file = find_file(root_dir, "*log.txt")
    except:
        raise ValueError("Experiment log file not found.")
    with open(log_file, "r") as f:
        log_parsed = log_parser.parse(f)

    acq_parser = Parser(acq_grammar)
    try:
        acq_file = find_file(root_dir, "*[Aa]cq.txt")
    except:
        raise ValueError("Experiment acq file not found.")
    with open(acq_file, "r") as f:
        acq_parsed = acq_parser.parse(f)

    parsed = {**acq_parsed, **log_parsed}

    for key, value in parsed.items():
        if isinstance(value, datetime):
            parsed[key] = datetime_to_timestamp(value)

    parsed_flattened = flatten_dict(parsed)
    for k, v in parsed_flattened.items():
        if isinstance(v, list):
            parsed_flattened[k] = [0 if el is None else el for el in v]

    return parsed_flattened



"""Read and convert MATLAB files from Swain Lab platform.

TODO: Information that I need from lab members esp J and A
    * Lots of examples to try
    * Any ideas on what these Map objects are?

TODO: Update Swain Lab wiki

All credit to Matt Bauman for
the reverse engineering at https://nbviewer.jupyter.org/gist/mbauman/9121961
"""

import re
import struct
import sys
from collections import Iterable
from io import BytesIO

import h5py
import numpy as np
import pandas as pd
import scipy
from numpy.compat import asstr

# TODO only use this if scipy>=1.6 or so
from scipy.io import matlab
from scipy.io.matlab.mio5 import MatFile5Reader
from scipy.io.matlab.mio5_params import mat_struct

from aliby.io.utils import read_int, read_string, read_delim


def read_minimat_vars(rdr):
    rdr.initialize_read()
    mdict = {"__globals__": []}
    i = 0
    while not rdr.end_of_stream():
        hdr, next_position = rdr.read_var_header()
        name = asstr(hdr.name)
        if name == "":
            name = "var_%d" % i
            i += 1
        res = rdr.read_var_array(hdr, process=False)
        rdr.mat_stream.seek(next_position)
        mdict[name] = res
        if hdr.is_global:
            mdict["__globals__"].append(name)
    return mdict


def read_workspace_vars(fname):
    fp = open(fname, "rb")
    rdr = MatFile5Reader(fp, struct_as_record=True, squeeze_me=True)
    vars = rdr.get_variables()
    fws = vars["__function_workspace__"]
    ws_bs = BytesIO(fws.tostring())
    ws_bs.seek(2)
    rdr.mat_stream = ws_bs
    # Guess byte order.
    mi = rdr.mat_stream.read(2)
    rdr.byte_order = mi == b"IM" and "<" or ">"
    rdr.mat_stream.read(4)  # presumably byte padding
    mdict = read_minimat_vars(rdr)
    fp.close()
    return mdict


class matObject:
    """A python read-out of MATLAB objects
    The objects pulled out of the
    """

    def __init__(self, filepath):
        self.filepath = filepath  # For record
        self.classname = None
        self.object_name = None
        self.buffer = None
        self.version = None
        self.names = None
        self.segments = None
        self.heap = None
        self.attrs = dict()
        self._init_buffer()
        self._init_heap()
        self._read_header()
        self.parse_file()

    def __getitem__(self, item):
        return self.attrs[item]

    def keys(self):
        """Returns the names of the available properties"""
        return self.attrs.keys()

    def get(self, item, default=None):
        return self.attrs.get(item, default)

    def _init_buffer(self):
        fp = open(self.filepath, "rb")
        rdr = MatFile5Reader(fp, struct_as_record=True, squeeze_me=True)
        vars = rdr.get_variables()
        self.classname = vars["None"]["s2"][0].decode("utf-8")
        self.object_name = vars["None"]["s0"][0].decode("utf-8")
        fws = vars["__function_workspace__"]
        self.buffer = BytesIO(fws.tostring())
        fp.close()

    def _init_heap(self):
        super_data = read_workspace_vars(self.filepath)
        elem = super_data["var_0"][0, 0]
        if isinstance(elem, mat_struct):
            self.heap = elem.MCOS[0]["arr"]
        else:
            self.heap = elem["MCOS"][0]["arr"]

    def _read_header(self):
        self.buffer.seek(248)  # the start of the header
        version = read_int(self.buffer)
        n_str = read_int(self.buffer)

        offsets = read_int(self.buffer, n=6)

        # check that the next two are zeros
        reserved = read_int(self.buffer, n=2)
        assert all(
            [x == 0 for x in reserved]
        ), "Non-zero reserved header fields: {}".format(reserved)
        # check that we are at the right place
        assert self.buffer.tell() == 288, "String elemnts begin at 288"
        hdrs = []
        for i in range(n_str):
            hdrs.append(read_string(self.buffer))
        self.names = hdrs
        self.version = version
        # The offsets are actually STARTING FROM 248 as well
        self.segments = [x + 248 for x in offsets]  # list(offsets)
        return

    def parse_file(self):
        # Get class attributes from segment 1
        self.buffer.seek(self.segments[0])
        classes = self._parse_class_attributes(self.segments[1])
        # Get first set of properties from segment 2
        self.buffer.seek(self.segments[1])
        props1 = self._parse_properties(self.segments[2])
        # Get the property description from segment 3
        self.buffer.seek(self.segments[2])
        object_info = self._parse_prop_description(classes, self.segments[3])
        # Get more properties from segment 4
        self.buffer.seek(self.segments[3])
        props2 = self._parse_properties(self.segments[4])
        # Check that the last segment is empty
        self.buffer.seek(self.segments[4])
        seg5_length = (self.segments[5] - self.segments[4]) // 8
        read_delim(self.buffer, seg5_length)
        props = (props1, props2)
        self._to_attrs(object_info, props)

    def _to_attrs(self, object_info, props):
        """Re-organise the various classes and subclasses into a nested
        dictionary.
        :return:
        """
        for pkg_clss, indices, idx in object_info:
            pkg, clss = pkg_clss
            idx = max(indices)
            which = indices.index(idx)
            obj = flatten_obj(props[which][idx])
            subdict = self.attrs
            if pkg != "":
                subdict = self.attrs.setdefault(pkg, {})
            if clss in subdict:
                if isinstance(subdict[clss], list):
                    subdict[clss].append(obj)
                else:
                    subdict[clss] = [subdict[clss]]
                    subdict[clss].append(obj)
            else:
                subdict[clss] = obj

    def describe(self):
        describe(self.attrs)

    def _parse_class_attributes(self, section_end):
        """Read the Class attributes = the first segment"""
        read_delim(self.buffer, 4)
        classes = []
        while self.buffer.tell() < section_end:
            package_index = read_int(self.buffer) - 1
            package = self.names[package_index] if package_index > 0 else ""
            name_idx = read_int(self.buffer) - 1
            name = self.names[name_idx] if name_idx > 0 else ""
            classes.append((package, name))
            read_delim(self.buffer, 2)
        return classes

    def _parse_prop_description(self, classes, section_end):
        """Parse the description of each property = the third segment"""
        read_delim(self.buffer, 6)
        object_info = []
        while self.buffer.tell() < section_end:
            class_idx = read_int(self.buffer) - 1
            class_type = classes[class_idx]
            read_delim(self.buffer, 2)
            indices = [x - 1 for x in read_int(self.buffer, 2)]
            obj_id = read_int(self.buffer)
            object_info.append((class_type, indices, obj_id))
        return object_info

    def _parse_properties(self, section_end):
        """
        Parse the actual values of the attributes == segments 2 and 4
        """
        read_delim(self.buffer, 2)
        props = []
        while self.buffer.tell() < section_end:
            n_props = read_int(self.buffer)
            d = parse_prop(n_props, self.buffer, self.names, self.heap)
            if not d:  # Empty dictionary
                break
            props.append(d)
            # Move to next 8-byte aligned offset
            self.buffer.seek(self.buffer.tell() + self.buffer.tell() % 8)
        return props

    def to_hdf(self, filename):
        f = h5py.File(filename, mode="w")
        save_to_hdf(f, "/", self.attrs)


def describe(d, indent=0, width=4, out=None):
    for key, value in d.items():
        print(f'{"": <{width * indent}}' + str(key), file=out)
        if isinstance(value, dict):
            describe(value, indent + 1, out=out)
        elif isinstance(value, np.ndarray):
            print(
                f'{"": <{width * (indent + 1)}} {value.shape} array '
                f"of type {value.dtype}",
                file=out,
            )
        elif isinstance(value, scipy.sparse.csc.csc_matrix):
            print(
                f'{"": <{width * (indent + 1)}} {value.shape} '
                f"sparse matrix of type {value.dtype}",
                file=out,
            )
        elif isinstance(value, Iterable) and not isinstance(value, str):
            print(
                f'{"": <{width * (indent + 1)}} {type(value)} of len ' f"{len(value)}",
                file=out,
            )
        else:
            print(f'{"": <{width * (indent + 1)}} {value}', file=out)


def parse_prop(n_props, buff, names, heap):
    d = dict()
    for i in range(n_props):
        name_idx, flag, heap_idx = read_int(buff, 3)
        if flag not in [0, 1, 2] and name_idx == 0:
            n_props = flag
            buff.seek(buff.tell() - 1)  # go back on one byte
            d = parse_prop(n_props, buff, names, heap)
        else:
            item_name = names[name_idx - 1]
            if flag == 0:
                d[item_name] = names[heap_idx]
            elif flag == 1:
                d[item_name] = heap[heap_idx + 2]  # Todo: what is the heap?
            elif flag == 2:
                assert 0 <= heap_idx <= 1, (
                    "Boolean flag has a value other " "than 0 or 1 "
                )
                d[item_name] = bool(heap_idx)
            else:
                raise ValueError(
                    "unknown flag {} for property {} with heap "
                    "index {}".format(flag, item_name, heap_idx)
                )
    return d


def is_object(x):
    """Checking object dtype for structured numpy arrays"""
    if x.dtype.names is not None and len(x.dtype.names) > 1:  # Complex obj
        return all(x.dtype[ix] == np.object for ix in range(len(x.dtype)))
    else:  # simple object
        return x.dtype == np.object


def flatten_obj(arr):
    # TODO turn structured arrays into nested dicts of lists rather that
    #  lists of dicts
    if isinstance(arr, np.ndarray):
        if arr.dtype.names:
            arrdict = dict()
            for fieldname in arr.dtype.names:
                arrdict[fieldname] = flatten_obj(arr[fieldname])
            arr = arrdict
        elif arr.dtype == np.object and arr.ndim == 0:
            arr = flatten_obj(arr[()])
        elif arr.dtype == np.object and arr.ndim > 0:
            try:
                arr = np.stack(arr)
                if arr.dtype.names:
                    d = {k: flatten_obj(arr[k]) for k in arr.dtype.names}
                    arr = d
            except:
                arr = [flatten_obj(x) for x in arr.tolist()]
    elif isinstance(arr, dict):
        arr = {k: flatten_obj(v) for k, v in arr.items()}
    elif isinstance(arr, list):
        try:
            arr = flatten_obj(np.stack(arr))
        except:
            arr = [flatten_obj(x) for x in arr]
    return arr


def save_to_hdf(h5file, path, dic):
    """
    Saving a MATLAB object to HDF5
    """
    if isinstance(dic, list):
        dic = {str(i): v for i, v in enumerate(dic)}
    for key, item in dic.items():
        if isinstance(item, (int, float, str)):
            h5file[path].attrs.create(key, item)
        elif isinstance(item, list):
            if len(item) == 0 and path + key not in h5file:  # empty list empty group
                h5file.create_group(path + key)
            if all(isinstance(x, (int, float, str)) for x in item):
                if path not in h5file:
                    h5file.create_group(path)
                h5file[path].attrs.create(key, item)
            else:
                if path + key not in h5file:
                    h5file.create_group(path + key)
                save_to_hdf(
                    h5file, path + key + "/", {str(i): x for i, x in enumerate(item)}
                )
        elif isinstance(item, scipy.sparse.csc.csc_matrix):
            try:
                h5file.create_dataset(
                    path + key, data=item.todense(), compression="gzip"
                )
            except Exception as e:
                print(path + key)
                raise e
        elif isinstance(item, (np.ndarray, np.int64, np.float64)):
            if item.dtype == np.dtype("<U1"):  # Strings to 'S' type for HDF5
                item = item.astype("S")
            try:
                h5file.create_dataset(path + key, data=item, compression="gzip")
            except Exception as e:
                print(path + key)
                raise e
        elif isinstance(item, dict):
            if path + key not in h5file:
                h5file.create_group(path + key)
            save_to_hdf(h5file, path + key + "/", item)
        elif item is None:
            continue
        else:
            raise ValueError(f"Cannot save {type(item)} type at key {path + key}")


## NOT YET FULLY IMPLEMENTED!


class _Info:
    def __init__(self, info):
        self.info = info
        self._identity = None

    def __getitem__(self, item):
        val = self.info[item]
        if val.shape[0] == 1:
            val = val[0]
        if 0 in val[1].shape:
            val = val[0]
        if isinstance(val, scipy.sparse.csc.csc_matrix):
            return np.asarray(val.todense())
        if val.dtype == np.dtype("O"):
            # 3d "sparse matrix"
            if all(isinstance(x, scipy.sparse.csc.csc_matrix) for x in val):
                val = np.array([x.todense() for x in val])
            # TODO: The actual object data
        equality = val[0] == val[1]
        if isinstance(equality, scipy.sparse.csc.csc_matrix):
            equality = equality.todense()
        if equality.all():
            val = val[0]
        return np.squeeze(val)

    @property
    def categories(self):
        return self.info.dtype.names


class TrapInfo(_Info):
    def __init__(self, info):
        """
        The information on all of the traps in a given position.

        :param info: The TrapInfo structure, can be found in the heap of
        the CTimelapse at index 7
        """
        super().__init__(info)


class CellInfo(_Info):
    def __init__(self, info):
        """
        The extracted information of all cells in a given position.
        :param info: The CellInfo structure, can be found in the heap
        of the CTimelapse at index 15.
        """
        super().__init__(info)

    @property
    def identity(self):
        if self._identity is None:
            self._identity = pd.DataFrame(
                zip(self["trapNum"], self["cellNum"]), columns=["trapNum", "cellNum"]
            )
        return self._identity

    def index(self, trapNum, cellNum):
        query = "trapNum=={} and cellNum=={}".format(trapNum, cellNum)
        try:
            result = self.identity.query(query).index[0]
        except Exception as e:
            print(query)
            raise e
        return result

    @property
    def nucEstConv1(self):
        return np.asarray(self.info["nuc_est_conv"][0][0].todense())

    @property
    def nucEstConv2(self):
        return np.asarray(self.info["nuc_est_conv"][0][1].todense())

    @property
    def mothers(self):
        return np.where((self["births"] != 0).any(axis=1))[0]

    def daughters(self, mother_index):
        """
        Get daughters of cell with index `mother_index`.

        :param mother_index: the index of the mother within the data. This is
        different from the mother's cell/trap identity.
        """
        daughter_ids = np.unique(self["daughterLabel"][mother_index]).tolist()
        daughter_ids.remove(0)
        mother_trap = self.identity["trapNum"].loc[mother_index]
        daughters = [self.index(mother_trap, cellNum) for cellNum in daughter_ids]
        return daughters


def _todict(matobj):
    """
    A recursive function which constructs from matobjects nested dictionaries
    """
    if not hasattr(matobj, "_fieldnames"):
        return matobj
    d = {}
    for strg in matobj._fieldnames:
        elem = matobj.__dict__[strg]
        if isinstance(elem, matlab.mio5_params.mat_struct):
            d[strg] = _todict(elem)
        elif isinstance(elem, np.ndarray):
            d[strg] = _toarray(elem)
        else:
            d[strg] = elem
    return d


def _toarray(ndarray):
    """
    A recursive function which constructs ndarray from cellarrays
    (which are loaded as numpy ndarrays), recursing into the elements
    if they contain matobjects.
    """
    if ndarray.dtype != "float64":
        elem_list = []
        for sub_elem in ndarray:
            if isinstance(sub_elem, matlab.mio5_params.mat_struct):
                elem_list.append(_todict(sub_elem))
            elif isinstance(sub_elem, np.ndarray):
                elem_list.append(_toarray(sub_elem))
            else:
                elem_list.append(sub_elem)
        return np.array(elem_list)
    else:
        return ndarray


from pathlib import Path


class Strain:
    """The cell info for all the positions of a strain."""

    def __init__(self, origin, strain):
        self.origin = Path(origin)
        self.files = [x for x in origin.iterdir() if strain in str(x)]
        self.cts = [matObject(x) for x in self.files]
        self.cinfos = [CellInfo(x.heap[15]) for x in self.cts]
        self._identity = None

    def __getitem__(self, item):
        try:
            return np.concatenate([c[item] for c in self.cinfos])
        except ValueError:  # If first axis is the channel
            return np.concatenate([c[item] for c in self.cinfos], axis=1)

    @property
    def categories(self):
        return set.union(*[set(c.categories) for c in self.cinfos])

    @property
    def identity(self):
        if self._identity is None:
            identities = []
            for pos_id, cinfo in enumerate(self.cinfos):
                identity = cinfo.identity
                identity["position"] = pos_id
                identities.append(identity)
            self._identity = pd.concat(identities, ignore_index=True)
        return self._identity

    def index(self, posNum, trapNum, cellNum):
        query = "position=={} and trapNum=={} and cellNum=={}".format(
            posNum, trapNum, cellNum
        )
        try:
            result = self.identity.query(query).index[0]
        except Exception as e:
            raise e
        return result

    @property
    def mothers(self):
        # At least two births are needed to be considered a mother cell
        return np.where(np.count_nonzero(self["births"], axis=1) > 3)[0]

    def daughters(self, mother_index):
        """
        Get daughters of cell with index `mother_index`.

        :param mother_index: the index of the mother within the data. This is
        different from the mother's pos/trap/cell identity.
        """
        daughter_ids = np.unique(self["daughterLabel"][mother_index]).tolist()
        if 0 in daughter_ids:
            daughter_ids.remove(0)
        mother_pos_trap = self.identity[["position", "trapNum"]].loc[mother_index]
        daughters = []
        for cellNum in daughter_ids:
            try:
                daughters.append(self.index(*mother_pos_trap, cellNum))
            except IndexError:
                continue
        return daughters



from typing import Union
import collections
from itertools import groupby, chain, product

import numpy as np
import h5py


class BridgeH5:
    """
    Base class to interact with h5 data stores.
    It also contains functions useful to predict how long should segmentation take.
    """

    def __init__(self, filename, flag="r"):
        self.filename = filename
        if flag is not None:
            self._hdf = h5py.File(filename, flag)

            self._filecheck

    def _filecheck(self):
        assert "cell_info" in self._hdf, "Invalid file. No 'cell_info' found."

    def close(self):
        self._hdf.close()

    def max_ncellpairs(self, nstepsback):
        """
        Get maximum number of cell pairs to be calculated
        """

        dset = self._hdf["cell_info"][()]
        # attrs = self._hdf[dataset].attrs
        pass

    @property
    def cell_tree(self):
        return self.get_info_tree()

    def get_n_cellpairs(self, nstepsback=2):
        cell_tree = self.cell_tree
        # get pair of consecutive trap-time points
        pass

    @staticmethod
    def get_consecutives(tree, nstepsback):
        # Receives a sorted tree and returns the keys of consecutive elements
        vals = {k: np.array(list(v)) for k, v in tree.items()}  # get tp level
        where_consec = [
            {
                k: np.where(np.subtract(v[n + 1 :], v[: -n - 1]) == n + 1)[0]
                for k, v in vals.items()
            }
            for n in range(nstepsback)
        ]  # get indices of consecutive elements
        return where_consec

    def get_npairs(self, nstepsback=2, tree=None):
        if tree is None:
            tree = self.cell_tree

        consecutive = self.get_consecutives(tree, nstepsback=nstepsback)
        flat_tree = flatten(tree)

        n_predictions = 0
        for i, d in enumerate(consecutive, 1):
            flat = list(chain(*[product([k], list(v)) for k, v in d.items()]))
            pairs = [(f, (f[0], f[1] + i)) for f in flat]
            for p in pairs:
                n_predictions += len(flat_tree.get(p[0], [])) * len(
                    flat_tree.get(p[1], [])
                )

        return n_predictions

    def get_npairs_over_time(self, nstepsback=2):
        tree = self.cell_tree
        npairs = []
        for t in self._hdf["cell_info"]["processed_timepoints"][()]:
            tmp_tree = {
                k: {k2: v2 for k2, v2 in v.items() if k2 <= t} for k, v in tree.items()
            }
            npairs.append(self.get_npairs(tree=tmp_tree))

        return np.diff(npairs)

    def get_info_tree(
        self, fields: Union[tuple, list] = ("trap", "timepoint", "cell_label")
    ):
        """
        Returns traps, time points and labels for this position in form of a tree
        in the hierarchy determined by the argument fields. Note that it is
        compressed to non-empty elements and timepoints.

        Default hierarchy is:
        - trap
         - time point
          - cell label

        This function currently produces trees of depth 3, but it can easily be
        extended for deeper trees if needed (e.g. considering groups,
        chambers and/or positions).

        input
        :fields: Fields to fetch from 'cell_info' inside the hdf5 storage

        returns
        :tree: Nested dictionary where keys (or branches) are the upper levels
             and the leaves are the last element of :fields:.
        """
        zipped_info = (*zip(*[self._hdf["cell_info"][f][()] for f in fields]),)

        return recursive_groupsort(zipped_info)


def groupsort(iterable: Union[tuple, list]):
    # Sorts iterable and returns a dictionary where the values are grouped by the first element.

    iterable = sorted(iterable, key=lambda x: x[0])
    grouped = {k: [x[1:] for x in v] for k, v in groupby(iterable, lambda x: x[0])}
    return grouped


def recursive_groupsort(iterable):
    # Recursive extension of groupsort
    if len(iterable[0]) > 1:
        return {k: recursive_groupsort(v) for k, v in groupsort(iterable).items()}
    else:  # Only two elements in list
        return [x[0] for x in iterable]


def flatten(d, parent_key="", sep="_"):
    """Flatten nested dict. Adapted from https://stackoverflow.com/a/6027615"""
    items = []
    for k, v in d.items():
        new_key = parent_key + (k,) if parent_key else (k,)
        if isinstance(v, collections.MutableMapping):
            items.extend(flatten(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)



"""
Utility functions and classes
"""
import itertools
import logging
import operator
from pathlib import Path
from typing import Callable

import h5py
import imageio
import cv2
import numpy as np

def repr_obj(obj, indent=0):
    """
    Helper function to display info about OMERO objects.
    Not all objects will have a "name" or owner field.
    """
    string = """%s%s:%s  Name:"%s" (owner=%s)""" % (
        " " * indent,
        obj.OMERO_CLASS,
        obj.getId(),
        obj.getName(),
        obj.getAnnotation())

    return string

def imread(path):
    return cv2.imread(str(path), -1)


class ImageCache:
    """HDF5-based image cache for faster loading of the images once they've
    been read.
    """
    def __init__(self, file, name, shape, remote_fn):
        self.store = h5py.File(file, 'a')
        # Create a dataset
        self.dataset = self.store.create_dataset(name, shape,
                                                 dtype=np.float,
                                                 fill_value=np.nan)
        self.remote_fn = remote_fn

    def __getitem__(self, item):
        cached = self.dataset[item]
        if np.any(np.isnan(cached)):
            full = self.remote_fn(item)
            self.dataset[item] = full
            return full
        else:
            return cached


class Cache:
    """
    Fixed-length mapping to use as a cache.
    Deletes items in FIFO manner when maximum allowed length is reached.
    """
    def __init__(self, max_len=5000, load_fn: Callable = imread):
        """
        :param max_len: Maximum number of items in the cache.
        :param load_fn: The function used to load new items if they are not
        available in the Cache
        """
        self._dict = dict()
        self._queue = []
        self.load_fn = load_fn
        self.max_len=max_len

    def __getitem__(self, item):
        if item not in self._dict:
            self.load_item(item)
        return self._dict[item]

    def load_item(self, item):
        self._dict[item] = self.load_fn(item)
        # Clean up the queue
        self._queue.append(item)
        if len(self._queue) > self.max_len:
            del self._dict[self._queue.pop(0)]

    def clear(self):
        self._dict.clear()
        self._queue.clear()


def accumulate(l: list):
    l = sorted(l)
    it = itertools.groupby(l, operator.itemgetter(0))
    for key, sub_iter in it:
        yield key, [x[1] for x in sub_iter]


def get_store_path(save_dir, store, name):
    """Create a path to a position-specific store.

    This combines the name and the store's base name into a file path within save_dir.
    For example.
    >>> get_store_path('data', 'baby_seg.h5', 'pos001')
    Path(data/pos001baby_seg.h5')

    :param save_dir: The root directory in which to save the file, absolute
    path.
    :param store: The base name of the store
    :param name: The name of the position
    :return: Path(save_dir) / name+store
    """
    store = Path(save_dir) / store
    store = store.with_name(name + store.name)
    return store

def parametrized(dec):
    def layer(*args, **kwargs):
        def repl(f):
            return dec(f, *args, **kwargs)
        return repl
    return layer

from functools import wraps, partial
from time import perf_counter
import logging
@parametrized
def timed(f, name=None):
    @wraps(f)
    def decorated(*args, **kwargs):
        t = perf_counter()
        res = f(*args, **kwargs)
        to_print = name or f.__name__
        logging.debug(f'Timing:{to_print}:{perf_counter() - t}s')
        return res
    return decorated






from pathos.multiprocessing import Pool

from aliby.pipeline import PipelineParameters, Pipeline


class MultiExp:
    """
    Manages cases when you need to segment several different experiments with a single
    position (e.g. pH calibration).
    """

    def __init__(self, expt_ids, npools=8, *args, **kwargs):

        self.expt_ids = expt_ids

    def run(self):
        run_expt = lambda expt: Pipeline(
            PipelineParameters.default(general={"expt_id": expt, "distributed": 0})
        ).run()
        with Pool(npools) as p:
            results = p.map(lambda x: self.create_pipeline(x), self.exp_ids)

    @classmethod
    def default(self):
        return cls(expt_ids=list(range(20448, 20467 + 1)))
